/*
 * leds
 * Linux Editor for D
 * S as in simple(superfluous) or supperb
 * 
 * by Antonio Monteiro
 */

module leds.CodeView;

private import dui.All;

private import std.c.stdio;
private import std.string;
private import std.path;
private alias std.path filePath;

private import std.file;
private import std.conv;
//private import std.syserror;
private import std.outbuffer;

private import leds.Leds;
private import leds.Project; 
private import leds.Workspace; 
private import leds.SystemConsole;
private import leds.POpen;
private import leds.FileView;
private import leds.Scintilla;
private import leds.FindAndReplace;
private import leds.TipWindow;

/**
 * A scrolled window that contains a scintilla widget
 */
public:
class CodeView : FileView , ScintillaListener , POpenLineListener
{
	
	protected:

	Project project;

	Scintilla scintilla;

	char[] previousFindString;

	char[][int] errors;

	bit autoIndent = true;

	public:
	
	/**
	 * creates a new empty CodeView
	 */
	this(Project project, Workspace workspace)
	{
		this(project, workspace, null);
	}

	/**
	 * creates a new CodeView and reads in a file
	 * @param fileName the fileName
	 */
	this(Project project, Workspace workspace, char[] fileName)
	{
		super(workspace, fileName);
		this.project = project;

		previousFindString = "";

		scintilla = new Scintilla();
		//scintilla.setRequest(-1,750);
		scintilla.setDLex();

		// this will emite a message to the console:
		// cannot add non scrollable widget use gtk_scrolled_window_add_with_viewport() instead
		// we need to ignore that
		add(scintilla);
		setupNotifications();
		if ( fileName !== null )
		{
			loadFile(fileName);
		}
	}

	void setBrowserView()
	{
		if ( Leds.leds.getPropValue("showCodeBrowser",false) )
		{
			printf("CodeView.setBrowserView 1\n");
			if ( workspace !== null )
			{
				printf("CodeView.setBrowserView 2\n");
				if ( filePath.getExt(getFileName()) == "d" )
				{
					printf("CodeView.setBrowserView 3\n");
					workspace.getBrowserView().parse(scintilla.getText());
				}
			}
		}
	}

	//~this()
	//{
	//	scintilla.release();
	//}

	/**
	 * loads a file into the view
	 * @param fileName the file to load
	 * @return true if sucesseful
	 */
	bit loadFile(char[] fileName)
	{
		this.fileName = fileName.dup;
		char[] input;
		try
		{
			input = cast(char[])std.file.read(fileName);
			scintilla.send(SCI_STYLECLEARALL, 0, 0);
			setLexMode(fileName);
			scintilla.send(SCI_INSERTTEXT, 0, input);
			scintilla.send(SCI_EMPTYUNDOBUFFER,0 ,0 );
			setBrowserView();
		}
		catch ( Error e)
		{
			e.print();
		}
		return true;
	}

	void setLexMode(char[] fileName)
	{
		if ( fileName.length>=8 && 0==std.string.icmp("makefile",fileName[0..7]) )
		{
			scintilla.setMakefileLex();
		}
		else
		{
			switch ( filePath.getExt(fileName) )
			{
			
				case "HTML":
				case "html":
				case "HTM":
				case "htm":
					scintilla.setHTMLLex();
					break;
					
				default:
					scintilla.setDLex();
					break;
			}
		}

	}

	/**
	 * Indicates if the text was modified
	 * @return true if the text was modified
	 */
	bit getModified()
	{
		return false;
	}
	
	/**
	 * indicates if this buffer is a empty never used buffer
	 * @return true if this instance should be loaded with a new file
	 */
	bit getEmpty()
	{
		return !getModified() && (fileName === null);
	}
	 
	/**
	 * checks if the buffer needs to be saved
	 * @return true if the buffer can be closed
	 */
	bit allowClose()
	{
		bit allow = true;
		if ( getModified() )
		{
			MessageDialog d = new MessageDialog(workspace, 0,
											DialogType.QUESTION,
											DialogButtons.NONE ,
											"File Not save!\nSave Now?");
			d.addButton(Stock.NO,Response.NO);
			d.addButton(Stock.YES,Response.YES);
			d.addButton(Stock.CANCEL,Response.CANCEL);
			int responce = d.run();
			if ( responce == Response.YES )
			{
				saveFile();
			}
			else if ( responce == Response.CANCEL )
			{
				allow = false;
			}
			d.destroy();
		}
		return allow;
	}

	/**
	 * saves the current file into the hard drive
	 * @return true if sucesseful
	 */
	bit saveFile()
	{
		OutBuffer buf = new OutBuffer();

		int textLength = scintilla.send(SCI_GETLENGTH,0,0)+1;
		
		char[] text = new char[textLength];
		char* p = &text[0];	
		scintilla.send(SCI_GETTEXT,textLength,(sptr_t)p);
		text.length = text.length-1;
		
		buf.write(text);
		try
		{
			.write(fileName,(byte[])buf.toBytes());
		}
		catch ( FileException fe )
		{
			printf("CodeView.saveFile FileException\n");
			workspace.saveAs(this, "Error! Please enter the file name");
		}
		//catch ( SysError se )
		//{
		//	printf("CodeView.saveFile SysError\n");
		//	workspace.saveAs(this, "Error! Please entre file name");
		//}
		catch ( Error e )
		{
			printf("CodeView.saveFile Error\n");
			workspace.saveAs(this, "Error! Please entre file name");
		}

		if ( filePath.getExt(getFileName()) == "d" )
		{
			checkSintax();
			setBrowserView();
		}

		return true;
	}

	/**
	 * Check the sintax by executing dmd to this file only
	 * @return true if the sintax is correct
	 */
	bit checkSintax()
	{
		bit correct = true;
		if ( project )
		{
			SystemConsole sc = workspace.getSystemConsole();
			char[] imports = ":" ~ Leds.leds().getPropValue("phobos:","~/devel/dmd/src/phobos");
			if ( project.getPropValue("type:","DUI") )
			{
				imports ~= ":" ~ Leds.leds().getPropValue("DUI:","~/devel/D/DUI/src");
			}
			
			char[][] otherImports = std.string.splitlines(project.getPropValue("importDirs:",""));
			for ( int i =0 ; i<otherImports.length ; i++ )
			{
				imports ~= ":" ~ otherImports[i];
			}

			scintilla.send(SCI_MARKERDELETEALL, SC_MARK_CIRCLE , 0);
			scintilla.send(SCI_MARKERSETBACK, SC_MARK_CIRCLE , 0x0000FF);
			errors = null ;
			new POpen(this
					,Leds.leds().getPropValue("dmd:","dmd")
					~ " " ~ getFileName()
					~ " -of/dev/null -c -I"
					~ project.getBaseDir()~"/src"
					~ imports
					~ " 2>&1"
					,POpenMode.READ
					,false);
		}
		return correct;
	}

	int getCodeLineNumber(inout char[] line)
	{
		int lineNumber = -1;
		int pos = std.string.find(line,"):");
		if ( pos > 1 )
		{
			int start = pos -1;
			while ( start >=0 && line[start] != '(' )
			{
				--start;
			}
			try
			{
				lineNumber = .toInt(line[start+1..pos]);
				if ( (pos+3) < line.length )
				{
					line = line[pos+3..line.length];
				}
			}
			catch ( Error e)
			{
				lineNumber = -1;
			}
		}
		return lineNumber;
	}

	bit popenLineCallback(char[] line)
	{
	printf("line callback %.*s\n",line);
		int lineNumber = getCodeLineNumber(line);
		if ( lineNumber > 0 )
		{
	printf("line callback 2\n");
			int markHandle = scintilla.send(SCI_MARKERADD, lineNumber-1, SC_MARK_CIRCLE);
			//errors[lineNumber] = line;
			errors[markHandle] = line;
		}
	printf("line callback 3\n");
		return false;
	}

	void popenClosed(char[] command)
	{
		printf("poppenClosed %.*s\n",command);
	}

	/**
	 * Copies the previous non white line leading blank characters
	 * into the negining of this line
	 */
	void doAutoIndent()
	{
		int lineNumber = scintilla.getCurrLine();
		int prevLine = lineNumber-1;
		while ( prevLine<=1 && scintilla.send(SCI_LINELENGTH, prevLine,0) )
		{
			--prevLine;
		}
		if ( prevLine >= 0 )
		{
			int indent = scintilla.send(SCI_GETLINEINDENTATION, lineNumber-1, 0);
			scintilla.send(SCI_SETLINEINDENTATION, lineNumber, indent);
			int newPos = scintilla.send(SCI_GETLINEINDENTPOSITION, lineNumber, 0);
			scintilla.send(SCI_SETCURRENTPOS, newPos, 0);
			scintilla.send(SCI_SETANCHOR, newPos, 0);
		}
	}

	/**
	 * renames and saves the current file into the storage device
	 * @return true if sucesseful
	 */
	bit saveFile(char[] fileName)
	{
		this.fileName = fileName.dup;
		return saveFile();
	}
	
	/**
	 * renames our current file name
	 */
	void rename(char[] fileName)
	{
	}

	/**
	 * execute an action if action is known.
	 * @param action
	 * @return true if the action was processed (even if the execution failled)
	 */
	public bit execAction(char[] action)
	{
		bit processed = true;
		switch(action)
		{
			case "view.zoom.in":	scintilla.send(SCI_ZOOMIN, 0, 0); break;
			case "view.zoom.out":   scintilla.send(SCI_ZOOMOUT, 0, 0); break;
			
			case "toolbar.undo":	scintilla.send(SCI_UNDO, 0, 0); break;
			case "toolbar.redo":	scintilla.send(SCI_REDO, 0, 0); break;

			case "toolbar.copy":	scintilla.send(SCI_COPY, 0, 0); break;
			case "toolbar.cut":	 	scintilla.send(SCI_CUT, 0, 0); break;
			case "toolbar.paste":	scintilla.send(SCI_PASTE, 0, 0); break;

			case "toolbar.findText": findText(); break;

			case "toolbar.gotoLine": 
				workspace.getGotoLine().show(scintilla.send(SCI_GETLINECOUNT, 0, 0));
				break;
				
			case "lexer.D": scintilla.setDLex();	break;
			case "lexer.makefile": scintilla.setMakefileLex();	break;
			case "lexer.html": scintilla.setMakefileLex();	break;
				
			default:
				processed = false;
			break;
		}
		return processed;
	}

	/**
	 * find for the text on the toolbar find text entry
	 */
	void findText()
	{
		findText(workspace.getFindText(),false);
	}

	/**
	 * find for a temporary text string
	 */
	void findText(char[] findStr, bit backwards)
	{
		if ( findStr === null || findStr == "" )
		{
			findStr = previousFindString;
		}
		if ( (findStr!==null) && (findStr.length>0) )
		{
			findReplace(findStr, null, false, true, 0, backwards);
		}
	}
	
	/**
	 * goto a line
	 */
	void gotoLine(int lineNumber)
	{
		if ( lineNumber > 0 )
		{
			scintilla.send(SCI_GOTOLINE, lineNumber-1, 0);
		}
	}
	
	/**
	 * Finds or replaces text
	 * @param findStr the string to find
	 * @param replaceStr the string to replace, the new string
	 * @param replace if true the replaceStr should replace the selected text
	 * @param find if true after replace do another find
	 * @param flags the find mode @see Scintilla
	 */
	void findReplace(char[] findStr, char[] replaceStr, bit replace, bit find,int flags, bit backwards)
	{
		
		if ( replace )
		{
			
		}
		
		if ( find )
		{
			previousFindString = findStr.dup;
			scintilla.find(findStr, flags, backwards);
		}
		
	}

	void setupNotifications()
	{
		scintilla.addScintillaListener((ScintillaListener)this);
		scintilla.send(SCN_KEY,0 ,0);
		scintilla.send(SCN_CHARADDED,0 ,0);
		scintilla.send(SCN_MARGINCLICK,0 ,0);
		scintilla.send(SCN_UPDATEUI,0 ,0);
	}
	
	/**
	 * Processes sintilla notifications
	 */
	void scintillaCallback(SCNotification* scn)
	{
		//printf("scinttilla callbale %d %d %d\n",scNotification.nmhdr.code,SCN_KEY,SCN_MARGINCLICK);
		switch(scn.nmhdr.code)
		{
			case SCN_UPDATEUI:
				int currPos = scintilla.getCurrPosition()-1;
				int other = scintilla.send(SCI_BRACEMATCH,currPos, 0);
				if ( other >= 0 )
				{
					scintilla.send(SCI_BRACEHIGHLIGHT,other, currPos);
					int column = scintilla.send(SCI_GETCOLUMN, currPos, 0);
					scintilla.send(SCI_SETHIGHLIGHTGUIDE, column, 0);
				}
				else
				{
					char c = scintilla.send(SCI_GETCHARAT, currPos, 0);
					if ( std.string.find("()[]{}",c) >= 0 )
					{
						scintilla.send(SCI_BRACEBADLIGHT, currPos, 0);
					}
					else
					{
						scintilla.send(SCI_BRACEBADLIGHT, -1, 0 );
					}
				}
				break;
				
			case SCN_CHARADDED:
				switch(scn.ch)
				{
					case 10:
						if ( autoIndent )
						{
							doAutoIndent();
						}
						break;

					default:
						break;
				}
			
			case SCN_KEY:
				if ( scn.modifiers == 2 )
				{
					switch ( scn.ch )
					{
						case 'S': workspace.saveCurrent();  break;
						case 'F': workspace.getFindAndReplace().show();	break;
						case 'W': workspace.closeView(); return;	break;
						case 'K': findText(scintilla.getSelected(),false); break;
						
						default:
							//printf("key %c %X \n",scn.ch,scn.modifiers);
							break;
					}
				}
				else if ( scn.modifiers == 3 )
				{
					switch ( scn.ch )
					{
						case 'S':
							workspace.saveCurrent();
							if ( project !== null )
							{
								project.runTool("makeCommand:", true);
							}
						break;
						case 'K': findText(scintilla.getSelected(),true); break;
						case 'P':
								int currPos = scintilla.getCurrPosition()-1;
								int other = scintilla.send(SCI_BRACEMATCH,currPos, 0);
								if ( other >= 0 )
								{
									scintilla.send(SCI_SETCURRENTPOS, other+1, 0);
									scintilla.send(SCI_SETANCHOR, other+1, 0);
								}
							break;
							
						default:
							//printf("key %c %X \n",scn.ch,scn.modifiers);
							break;
					}
				}
				break;
			
			case SCN_MARGINCLICK:
				
				int lineNumber = scintilla.send(SCI_LINEFROMPOSITION, scn.position,0);
				bit found = false;
				int i = 0;
				int[] keys = errors.keys;
				while(i<keys.length && !found)
				{
					int markerLine = scintilla.send(SCI_MARKERLINEFROMHANDLE, keys[i], 0 );
					found = lineNumber == markerLine;
					if ( !found) i++;
				}
				if ( found )
				{
					if ( keys[i] in errors )
					{
						workspace.getSystemConsole().append(errors[keys[i]]~\n);
						scintilla.send(SCI_CALLTIPSHOW, scn.position, errors[keys[i]]);
					}
				}
				break;
				
			default:
				//printf("%d %d CodeView.scintillaCallback\n",scn.ch,scn.nmhdr.code);
				break;
		}
		int pos = scintilla.send(SCI_GETCURRENTPOS,0,0);

		workspace.getStatusbar().setLinesCol(
			scintilla.send(SCI_LINEFROMPOSITION,pos , 0)+1,
			scintilla.send(SCI_GETLINECOUNT, 0, 0),
			scintilla.send(SCI_GETCOLUMN, pos, 0)
		);
		
	}

}
