/*
 * leds
 * Linux Editor for D 
 * S as in simple(superfluous) or supperb
 * 
 * by Antonio Monteiro
 */

module leds.FileView;

private import dui.All;

private import std.c.stdio;
private import std.string;
private import std.path;
private alias std.path filePath;

private import std.file;
private import std.outbuffer;

private import leds.Workspace;
private import leds.Docker;

/**
 * A scrolled window that contains a scintilla widget
 */
public:
class FileView : ScrolledWindow , Dockable
{

	protected:
	char[] fileName;
	Workspace workspace;
	Menu tabMenu;
	Docker docker;
	
	public:
	
	/**
	 * creates a new CodeView and reads in a file
	 * @param fileName the fileName
	 */
	this(Workspace workspace, char[] fileName)
	{
		super(PolicyType.AUTOMATIC, PolicyType.AUTOMATIC);
		this.workspace = workspace;
		this.fileName = fileName.dup;
	}

	Widget getWidget()
	{
		return this;
	}
	
	char[] getTabName()
	{
		if ( fileName === null || fileName.length == 0 )
		{
			return "Untitled";
		}
		return filePath.getBaseName(fileName);
	}

	Menu getTabMenu()
	{
		if ( tabMenu === null )
		{
			tabMenu = new Menu();
			MenuItem item;
			
			//tabMenu.append(new MenuItem(this,"_Close","tabMenu.close"));
			//tabMenu.append(new MenuItem(this,"_Float","tabMenu.float"));
			tabMenu.append(new MenuItem("test"));

		}
		//return tabMenu;
		return null;
	}

	void setDocker(Docker docker)
	{
		this.docker = docker;
	}

	/**
	 * Return this dockable file view - ourself
	 */
	FileView getFileView()
	{
		return this;
	}
	
	/**
	 * detriminis if a close button shoud be visible
	 */
	bit allowCloseButton()
	{
		return false;
	}
	
	bit allowClose()
	{
		return true;
	}

	bit clear()
	{
		return false;
	}

	/**
	 * Gets file name
	 * @return the file name
	 */
	char[] getFileName()
	{
		return fileName;
	}

	/**
	 * Gets the file name  without the full path
	 * @return the file name  without the full path
	 * or an emtpy string if the fileName was not yet set
	 */
	char[] getBaseName()
	{
		if ( fileName === null )
		{
			return "";
		}
		return filePath.getBaseName(fileName);
	}

	/**
	 * loads a file into the view
	 * @param fileName the file to load
	 * @return true if sucesseful
	 */
	abstract bit loadFile(char[] fileName);

	/**
	 * Indicates if the text was modified
	 * @return true if the text was modified
	 */
	abstract bit getModified();
	
	/**
	 * indicates if this buffer is a empty never used buffer
	 * @return true if this instance should be loaded with a new file
	 */
	abstract bit getEmpty();
	 
	/**
	 * saves the current file into the hard drive
	 * @return true if sucesseful
	 */
	abstract bit saveFile();
	
	/**
	 * renames and saves the current file into the hard drive
	 * @return true if sucesseful
	 */
	abstract bit saveFile(char[] fileName);
	
	/**
	 * renames our current file name
	 */
	abstract void rename(char[] fileName);

	/**
	 * execute an action if action is known.
	 * @param action
	 * @return true if the action was processed (even if the execution failled)
	 */
	abstract public bit execAction(char[] action);

	/**
	 * find for the text on the toolbar find text entry
	 */
	abstract void findText();
	
	/**
	 * goto a line
	 */
	abstract void gotoLine(int lineNumber);
	
	/**
	 * Finds or replaces text
	 * @param findStr the string to find
	 * @param replaceStr the string to replace, the new string
	 * @param replace if true the replaceStr should replace the selected text
	 * @param find if true after replace do another find
	 * @param flags the find mode @see Scintilla
	 */
	abstract void findReplace(char[] findStr, char[] replaceStr, bit replace, bit find,int flags, bit backwards);

}
